/* Copyright 1999-2003 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/serial.h>

#include "keyboard.h"

#ifdef __sparc__
#include <asm/openpromio.h>
#include <asm/types.h>
#endif

static void keyboardFreeDevice(struct keyboardDevice *dev)
{
	freeDevice((struct device *) dev);
}

static void keyboardWriteDevice(FILE *file, struct keyboardDevice *dev)
{
	writeDevice(file, (struct device *)dev);
}

static int keyboardCompareDevice(struct keyboardDevice *dev1, struct keyboardDevice *dev2)
{
	return compareDevice( (struct device *)dev1, (struct device *)dev2);
}

struct keyboardDevice *keyboardNewDevice(struct keyboardDevice *old)
{
	struct keyboardDevice *ret;

	ret = malloc(sizeof(struct keyboardDevice));
	memset(ret, '\0', sizeof(struct keyboardDevice));
	ret = (struct keyboardDevice *) newDevice((struct device *) old, (struct device *) ret);
	ret->bus = BUS_KEYBOARD;
	ret->newDevice = keyboardNewDevice;
	ret->freeDevice = keyboardFreeDevice;
	ret->writeDevice = keyboardWriteDevice;
	ret->compareDevice = keyboardCompareDevice;
	return ret;
}

struct speeds
{
	speed_t speed;
	unsigned long value;
};

struct speeds speed_map[] =
{
  {B50, 50},
  {B75, 75},
  {B110, 110},
  {B134, 134},
  {B150, 150},
  {B200, 200},
  {B300, 300},
  {B600, 600},
  {B1200, 1200},
  {B1800, 1800},
  {B2400, 2400},
  {B4800, 4800},
  {B9600, 9600},
  {B19200, 19200},
  {B38400, 38400},
  {B19200, 19200},
  {B38400, 38400},
#ifdef B57600
  {B57600, 57600},
#endif
#ifdef B115200
  {B115200, 115200},
#endif
#ifdef B230400
  {B230400, 230400},
#endif
#ifdef B460800
  {B460800, 460800},
#endif
  {0, 0}
};

#if !defined(__s390__) && !defined(__s390x__)
static int termcmp(struct termios *a, struct termios *b)
{
	if (a->c_iflag != b->c_iflag || a->c_oflag != b->c_oflag ||
	    a->c_cflag != b->c_cflag || a->c_lflag != b->c_lflag ||
	    cfgetispeed(a) != cfgetispeed(b) || cfgetospeed(a) != cfgetospeed(b))
		return 1;
	return memcmp(a->c_cc, b->c_cc, sizeof(a->c_cc));
}
#endif

struct device *keyboardProbe(enum deviceClass probeClass, int probeFlags,
			struct device *devlist)
{
	int fd;
	char twelve = 12;
	int fdstd = 0;
	char buf[4096];
	struct keyboardDevice *kbddev;
	
	if (probeClass & CLASS_KEYBOARD) {
#ifdef __powerpc__
		int cfd;
		struct termios cmode, mode;
		int x;

		cfd = open("/dev/console", O_RDONLY);
		tcgetattr(cfd,&cmode);
		close(cfd);

		fd = open("/dev/hvc0",O_RDONLY);
		if (fd >= 0 && !tcgetattr(fd, &mode) && !termcmp(&cmode, &mode)) {
			close(fd);
			kbddev=keyboardNewDevice(NULL);
			kbddev->type=CLASS_KEYBOARD;
			if (devlist)
				kbddev->next = devlist;
			devlist = (struct device *) kbddev;
			kbddev->desc=strdup("pSeries LPAR console");
			kbddev->device=strdup("hvc0");
		}
		for (x = 0; x < 2; x++) {
			sprintf(buf,"/dev/hvsi%d", x);
			fd = open(buf,O_RDONLY);
			if (fd >=0 && !tcgetattr(fd, &mode) && !termcmp(&cmode, &mode)) {
				close(fd);
				kbddev=keyboardNewDevice(NULL);
				kbddev->type = CLASS_KEYBOARD;
				if (devlist)
					kbddev->next = devlist;
				devlist = (struct device *) kbddev;
				kbddev->desc = malloc(20);
				snprintf(kbddev->desc,20,"HVSI Console %d",x);
				kbddev->device = malloc(10);
				sprintf(kbddev->device,"hvsi%d",x);
			}
		}
#endif
#ifdef __ia64__
		fd = open("/dev/ttySG0",O_RDONLY);
		if (fd >=0) {
			close(fd);
			kbddev = keyboardNewDevice(NULL);
			kbddev->type = CLASS_KEYBOARD;
			if (devlist)
				kbddev->next = devlist;
			devlist = (struct device *) kbddev;
			kbddev->desc = strdup("Altix Console");
			kbddev->device = strdup("ttySG0");
		}
#endif
#if defined(__i386__) || defined(__x86_64__) || defined(__ia64__) || defined(__sparc__)
		int cfd;
		struct termios cmode, mode;

		cfd = open("/dev/console", O_RDONLY);
		tcgetattr(cfd,&cmode);
		close(cfd);

                fd = open("/dev/xvc0", O_RDONLY);
		if (fd >= 0 && !tcgetattr(fd, &mode) && !termcmp(&cmode, &mode)) {
                    close(fd);
                    kbddev = keyboardNewDevice(NULL);
                    kbddev->type = CLASS_KEYBOARD;
                    if (devlist)
                        kbddev->next = devlist;
                    devlist = (struct device *) kbddev;
                    kbddev->desc = strdup("Xen Console");
                    kbddev->device = strdup("xvc0");
		}
#endif

		for (fd = 0; fd <= 2; fd++) {
		    sprintf (buf, "/proc/self/fd/%d", fd);
		    if (readlink (buf, buf, 4096) == 12 &&
			!strncmp (buf, "/dev/console", 12)) {
			fdstd = 1;
			break;
		    }
		}
		if (!fdstd) {
		    fd = open("/dev/console", O_RDWR);
		    if (fd < 0) goto out;
		}
		if (ioctl (fd, TIOCLINUX, &twelve) < 0) {
			/* Serial console */
			char desc[64];
			struct serial_struct si;
			struct termios tp;
			speed_t tspeed;
			int line = 0, speed = 0, i;

			if (ioctl (fd, TIOCGSERIAL, &si) >= 0) {
				int sfd;
				char path[64];
				struct serial_struct si2;
				
				line = si.line;
				sprintf(path,"/dev/ttyS%d",line);
				sfd = open(path, O_RDWR|O_NONBLOCK);
				if (sfd == -1)
					goto out;
				if (ioctl (sfd, TIOCGSERIAL, &si2) >= 0) {
					if (memcmp(&si,&si2,sizeof(si))) {
						close(sfd);
						goto out;
					}
				} else {
					goto out;
				}
				close(sfd);
			} else {
				/* It's not a serial port. Bail. */
				if (!fdstd) close(fd);
				goto out;
			}
			if (!tcgetattr(fd, &tp)) {
				tspeed = cfgetospeed(&tp);
				for (i = 0; speed_map[i].value != 0; i++)
					if (speed_map[i].speed == tspeed)
						speed = speed_map[i].value;
			}
			if (!fdstd) close(fd);
			kbddev=keyboardNewDevice(NULL);
			kbddev->type=CLASS_KEYBOARD;
			if (devlist)
				kbddev->next = devlist;
			devlist = (struct device *) kbddev;
			if (speed)
				sprintf (desc, "Serial console (%d baud) ttyS%d", speed, line);
			else
				sprintf (desc, "Serial console ttyS%d", line);
			kbddev->desc=strdup(desc);
			sprintf (desc, "ttyS%d", line);
			kbddev->device=strdup(desc);
		}
	}
out:
	return devlist;
}
